import 'dart:math';

import '../extensions/extensions.dart';
import '../renderer/actor.dart';
import '../renderer/actor_member.dart';
import '../renderer/anchor.dart';
import '../renderer/matrix_transformer.dart';

class RectPoints implements ActorMember {
  @override
  final Vector2 scale = Vector2.all(1);

  Vector2 topLeft;
  Vector2 topRight;
  Vector2 bottomLeft;
  Vector2 bottomRight;
  @override
  Anchor anchor;

  Vector2? _lateLineAngle;
  Vector2? _size;
  double? _angle;
  Vector2? _skew;
  Vector2? _position;

  RectPoints({
    required this.topLeft,
    required this.topRight,
    required this.bottomLeft,
    required this.bottomRight,
    this.anchor = Anchor.center,
  });

  RectPoints.fromActor(Actor actor)
      : this(
          anchor: actor.anchor,
          topLeft: actor.absolutePositionOfAnchor(Anchor.topLeft),
          topRight: actor.absolutePositionOfAnchor(Anchor.topRight),
          bottomLeft: actor.absolutePositionOfAnchor(Anchor.bottomLeft),
          bottomRight: actor.absolutePositionOfAnchor(Anchor.bottomRight),
        );

  void updateAnchor(Anchor anchor) {
    this.anchor = anchor;
    _position = null;
  }

  @override
  Vector2 get skew {
    if (_skew == null) {
      final horizontalLineAngle = _lineAngle.x;
      final verticalLineAngle = _lineAngle.y;

      double intersectionAngle = (horizontalLineAngle - verticalLineAngle).abs();
      if (intersectionAngle >= pi) {
        intersectionAngle = intersectionAngle - pi;
      }
      if (horizontalLineAngle >= 0 && verticalLineAngle < 0) {
        intersectionAngle = pi - intersectionAngle;
      }
      _skew = Vector2(pi / 2 - intersectionAngle, 0);
    }
    return _skew!;
  }

  @override
  Vector2 get position {
    if (_position == null) {
      final leftTopPosition = (topLeft + bottomRight) / 2 - size / 2;
      _position = leftTopPosition + offset;
    }
    return _position!;
  }

  @override
  double get angle {
    if (_angle == null) {
      double angle = _lineAngle.x;
      if (angle < 0) {
        angle += pi * 2;
      }
      _angle = angle;
    }
    return _angle!;
  }

  @override
  Vector2 get size {
    if (_size == null) {
      final size = Vector2(
        topRight.distanceTo(topLeft),
        bottomRight.distanceTo(topRight),
      );
      if (skew.x != 0) {
        size.y *= cos(skew.x.abs());
      }
      _size = size;
    }
    return _size!;
  }

  MatrixTransform get transform {
    return MatrixTransform(
      skew: skew,
      position: position,
      scale: scale,
      offset: -offset,
      angle: angle,
    );
  }

  Vector2 get offset => Vector2(anchor.x * size.x, anchor.y * size.y);

  Vector2 get _lineAngle {
    return _lateLineAngle ??= Vector2(
      atan2(topRight.y - topLeft.y, topRight.x - topLeft.x),
      atan2(bottomLeft.y - topLeft.y, bottomLeft.x - topLeft.x),
    );
  }
}
